@@ -30,7 +30,7 @@
 #include "dnswriter.hh"
 #include <unistd.h>
 
-bool g_verbose{true};
+bool g_verbose{false};
 bool g_syslog{true};
 
 BOOST_AUTO_TEST_SUITE(test_dnscrypt_cc)
@@ -49,20 +49,17 @@ BOOST_AUTO_TEST_CASE(DNSCryptPlaintextQuery) {
   auto ctx = std::make_shared<DNSCryptContext>("2.name", resolverCert, resolverPrivateKey);
 
   DNSName name("2.name.");
-  vector<uint8_t> plainQuery;
-  DNSPacketWriter pw(plainQuery, name, QType::TXT, QClass::IN, 0);
+  PacketBuffer plainQuery;
+  GenericDNSPacketWriter<PacketBuffer> pw(plainQuery, name, QType::TXT, QClass::IN, 0);
   pw.getHeader()->rd = 0;
-  uint16_t len = plainQuery.size();
 
   std::shared_ptr<DNSCryptQuery> query = std::make_shared<DNSCryptQuery>(ctx);
-  uint16_t decryptedLen = 0;
-
-  query->parsePacket((char*) plainQuery.data(), len, false, &decryptedLen, now);
+  query->parsePacket(plainQuery, false, now);
 
   BOOST_CHECK_EQUAL(query->isValid(), true);
   BOOST_CHECK_EQUAL(query->isEncrypted(), false);
 
-  std::vector<uint8_t> response;
+  PacketBuffer response;
 
   query->getCertificateResponse(now, response);
 
@@ -91,15 +88,12 @@ BOOST_AUTO_TEST_CASE(DNSCryptPlaintextQueryInvalidA) {
 
   DNSName name("2.name.");
 
-  vector<uint8_t> plainQuery;
-  DNSPacketWriter pw(plainQuery, name, QType::A, QClass::IN, 0);
+  PacketBuffer plainQuery;
+  GenericDNSPacketWriter<PacketBuffer> pw(plainQuery, name, QType::A, QClass::IN, 0);
   pw.getHeader()->rd = 0;
-  uint16_t len = plainQuery.size();
 
   std::shared_ptr<DNSCryptQuery> query = std::make_shared<DNSCryptQuery>(ctx);
-  uint16_t decryptedLen = 0;
-
-  query->parsePacket((char*) plainQuery.data(), len, false, &decryptedLen, now);
+  query->parsePacket(plainQuery, false, now);
 
   BOOST_CHECK_EQUAL(query->isValid(), false);
 }
@@ -117,15 +111,12 @@ BOOST_AUTO_TEST_CASE(DNSCryptPlaintextQueryInvalidProviderName) {
 
   DNSName name("2.WRONG.name.");
 
-  vector<uint8_t> plainQuery;
-  DNSPacketWriter pw(plainQuery, name, QType::TXT, QClass::IN, 0);
+  PacketBuffer plainQuery;
+  GenericDNSPacketWriter<PacketBuffer> pw(plainQuery, name, QType::TXT, QClass::IN, 0);
   pw.getHeader()->rd = 0;
-  uint16_t len = plainQuery.size();
 
   std::shared_ptr<DNSCryptQuery> query = std::make_shared<DNSCryptQuery>(ctx);
-  uint16_t decryptedLen = 0;
-
-  query->parsePacket((char*) plainQuery.data(), len, false, &decryptedLen, now);
+  query->parsePacket(plainQuery, false, now);
 
   BOOST_CHECK_EQUAL(query->isValid(), false);
 }
@@ -149,32 +140,24 @@ BOOST_AUTO_TEST_CASE(DNSCryptEncryptedQueryValid) {
   unsigned char clientNonce[DNSCRYPT_NONCE_SIZE / 2] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x08, 0x09, 0x0A, 0x0B };
 
   DNSName name("www.powerdns.com.");
-  vector<uint8_t> plainQuery;
-  DNSPacketWriter pw(plainQuery, name, QType::AAAA, QClass::IN, 0);
+  PacketBuffer plainQuery;
+  GenericDNSPacketWriter<PacketBuffer> pw(plainQuery, name, QType::AAAA, QClass::IN, 0);
   pw.getHeader()->rd = 1;
-  size_t requiredSize = plainQuery.size() + sizeof(DNSCryptQueryHeader) + DNSCRYPT_MAC_SIZE;
-  if (requiredSize < DNSCryptQuery::s_minUDPLength) {
-    requiredSize = DNSCryptQuery::s_minUDPLength;
-  }
 
-  plainQuery.reserve(requiredSize);
-  uint16_t len = plainQuery.size();
-  uint16_t encryptedResponseLen = 0;
-
-  int res = ctx->encryptQuery((char*) plainQuery.data(), len, plainQuery.capacity(), clientPublicKey, clientPrivateKey, clientNonce, false, &encryptedResponseLen, std::make_shared<DNSCryptCert>(resolverCert));
+  size_t initialSize = plainQuery.size();
+  int res = ctx->encryptQuery(plainQuery, 4096, clientPublicKey, clientPrivateKey, clientNonce, false, std::make_shared<DNSCryptCert>(resolverCert));
 
   BOOST_CHECK_EQUAL(res, 0);
-  BOOST_CHECK(encryptedResponseLen > len);
+  BOOST_CHECK(plainQuery.size() > initialSize);
 
   std::shared_ptr<DNSCryptQuery> query = std::make_shared<DNSCryptQuery>(ctx);
-  uint16_t decryptedLen = 0;
 
-  query->parsePacket((char*) plainQuery.data(), encryptedResponseLen, false, &decryptedLen, now);
+  query->parsePacket(plainQuery, false, now);
 
   BOOST_CHECK_EQUAL(query->isValid(), true);
   BOOST_CHECK_EQUAL(query->isEncrypted(), true);
 
-  MOADNSParser mdp(true, (char*) plainQuery.data(), decryptedLen);
+  MOADNSParser mdp(true, (char*) plainQuery.data(), plainQuery.size());
 
   BOOST_CHECK_EQUAL(mdp.d_header.qdcount, 1U);
   BOOST_CHECK_EQUAL(mdp.d_header.ancount, 0U);
@@ -205,15 +188,11 @@ BOOST_AUTO_TEST_CASE(DNSCryptEncryptedQueryValidButShort) {
   unsigned char clientNonce[DNSCRYPT_NONCE_SIZE / 2] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x08, 0x09, 0x0A, 0x0B };
 
   DNSName name("www.powerdns.com.");
-  vector<uint8_t> plainQuery;
-  DNSPacketWriter pw(plainQuery, name, QType::AAAA, QClass::IN, 0);
+  PacketBuffer plainQuery;
+  GenericDNSPacketWriter<PacketBuffer> pw(plainQuery, name, QType::AAAA, QClass::IN, 0);
   pw.getHeader()->rd = 1;
 
-  uint16_t len = plainQuery.size();
-  uint16_t encryptedResponseLen = 0;
-
-  int res = ctx->encryptQuery((char*) plainQuery.data(), len, plainQuery.capacity(), clientPublicKey, clientPrivateKey, clientNonce, false, &encryptedResponseLen, std::make_shared<DNSCryptCert>(resolverCert));
-
+  int res = ctx->encryptQuery(plainQuery, /* not enough room */ plainQuery.size(), clientPublicKey, clientPrivateKey, clientNonce, false, std::make_shared<DNSCryptCert>(resolverCert));
   BOOST_CHECK_EQUAL(res, ENOBUFS);
 }
 
@@ -236,24 +215,15 @@ BOOST_AUTO_TEST_CASE(DNSCryptEncryptedQueryValidWithOldKey) {
   unsigned char clientNonce[DNSCRYPT_NONCE_SIZE / 2] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x08, 0x09, 0x0A, 0x0B };
 
   DNSName name("www.powerdns.com.");
-  vector<uint8_t> plainQuery;
-  DNSPacketWriter pw(plainQuery, name, QType::AAAA, QClass::IN, 0);
+  PacketBuffer plainQuery;
+  GenericDNSPacketWriter<PacketBuffer> pw(plainQuery, name, QType::AAAA, QClass::IN, 0);
   pw.getHeader()->rd = 1;
 
-  size_t requiredSize = plainQuery.size() + sizeof(DNSCryptQueryHeader) + DNSCRYPT_MAC_SIZE;
-  if (requiredSize < DNSCryptQuery::s_minUDPLength) {
-    requiredSize = DNSCryptQuery::s_minUDPLength;
-  }
-
-  plainQuery.reserve(requiredSize);
-
-  uint16_t len = plainQuery.size();
-  uint16_t encryptedResponseLen = 0;
-
-  int res = ctx->encryptQuery((char*) plainQuery.data(), len, plainQuery.capacity(), clientPublicKey, clientPrivateKey, clientNonce, false, &encryptedResponseLen, std::make_shared<DNSCryptCert>(resolverCert));
+  size_t initialSize = plainQuery.size();
+  int res = ctx->encryptQuery(plainQuery, 4096, clientPublicKey, clientPrivateKey, clientNonce, false, std::make_shared<DNSCryptCert>(resolverCert));
 
   BOOST_CHECK_EQUAL(res, 0);
-  BOOST_CHECK(encryptedResponseLen > len);
+  BOOST_CHECK(plainQuery.size() > initialSize);
 
   DNSCryptCert newResolverCert;
   DNSCryptContext::generateCertificate(2, now, now + (24 * 60 * 3600), DNSCryptExchangeVersion::VERSION1, providerPrivateKey, resolverPrivateKey, newResolverCert);
@@ -261,14 +231,13 @@ BOOST_AUTO_TEST_CASE(DNSCryptEncryptedQueryValidWithOldKey) {
   ctx->markInactive(resolverCert.getSerial());
 
   std::shared_ptr<DNSCryptQuery> query = std::make_shared<DNSCryptQuery>(ctx);
-  uint16_t decryptedLen = 0;
 
-  query->parsePacket((char*) plainQuery.data(), encryptedResponseLen, false, &decryptedLen, now);
+  query->parsePacket(plainQuery, false, now);
 
   BOOST_CHECK_EQUAL(query->isValid(), true);
   BOOST_CHECK_EQUAL(query->isEncrypted(), true);
 
-  MOADNSParser mdp(true, (char*) plainQuery.data(), decryptedLen);
+  MOADNSParser mdp(true, (char*) plainQuery.data(), plainQuery.size());
 
   BOOST_CHECK_EQUAL(mdp.d_header.qdcount, 1U);
   BOOST_CHECK_EQUAL(mdp.d_header.ancount, 0U);
@@ -299,24 +268,15 @@ BOOST_AUTO_TEST_CASE(DNSCryptEncryptedQueryInvalidWithWrongKey) {
   unsigned char clientNonce[DNSCRYPT_NONCE_SIZE / 2] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x08, 0x09, 0x0A, 0x0B };
 
   DNSName name("www.powerdns.com.");
-  vector<uint8_t> plainQuery;
-  DNSPacketWriter pw(plainQuery, name, QType::AAAA, QClass::IN, 0);
+  PacketBuffer plainQuery;
+  GenericDNSPacketWriter<PacketBuffer> pw(plainQuery, name, QType::AAAA, QClass::IN, 0);
   pw.getHeader()->rd = 1;
 
-  size_t requiredSize = plainQuery.size() + sizeof(DNSCryptQueryHeader) + DNSCRYPT_MAC_SIZE;
-  if (requiredSize < DNSCryptQuery::s_minUDPLength) {
-    requiredSize = DNSCryptQuery::s_minUDPLength;
-  }
-
-  plainQuery.reserve(requiredSize);
-
-  uint16_t len = plainQuery.size();
-  uint16_t encryptedResponseLen = 0;
-
-  int res = ctx->encryptQuery((char*) plainQuery.data(), len, plainQuery.capacity(), clientPublicKey, clientPrivateKey, clientNonce, false, &encryptedResponseLen, std::make_shared<DNSCryptCert>(resolverCert));
+  size_t initialSize = plainQuery.size();
+  int res = ctx->encryptQuery(plainQuery, 4096, clientPublicKey, clientPrivateKey, clientNonce, false, std::make_shared<DNSCryptCert>(resolverCert));
 
   BOOST_CHECK_EQUAL(res, 0);
-  BOOST_CHECK(encryptedResponseLen > len);
+  BOOST_CHECK(plainQuery.size() > initialSize);
 
   DNSCryptCert newResolverCert;
   DNSCryptContext::generateCertificate(2, now, now + (24 * 60 * 3600), DNSCryptExchangeVersion::VERSION1, providerPrivateKey, resolverPrivateKey, newResolverCert);
@@ -327,9 +287,8 @@ BOOST_AUTO_TEST_CASE(DNSCryptEncryptedQueryInvalidWithWrongKey) {
   /* we have removed the old certificate, we can't decrypt this query */
 
   std::shared_ptr<DNSCryptQuery> query = std::make_shared<DNSCryptQuery>(ctx);
-  uint16_t decryptedLen = 0;
 
-  query->parsePacket((char*) plainQuery.data(), encryptedResponseLen, false, &decryptedLen, now);
+  query->parsePacket(plainQuery, false, now);
 
   BOOST_CHECK_EQUAL(query->isValid(), false);
 }
